// VinciaCommon.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for VinciaCommon.

#include "VinciaCommon.h"

namespace Vincia {

//**********************************************************************

// VINCIA Auxiliary helper functions

string num2str(int i, int width) {
      ostringstream tmp;
      if (width <= 1) tmp<<i;
      else if (abs(i) < pow(10.0,width-1)) {
	tmp<<fixed<<setw(width)<<i;
      } else {
	string ab="k";
	double r=i;
	if (abs(i) < 1e5) {
	  r/=1e3;
	} else if (abs(i) < 1e8) {
	  r/=1e6;
	  ab="M";
	} else if (abs(i) < 1e11) {
	  r/=1e9;
	  ab="G";
	} else if (abs(i) < 1e14) {
	  ab="T";
	  r/=1e12;
	}	
	tmp<<fixed<<setw(width-1)<<(r > 10 ? setprecision(width-4) : setprecision(width-3))
	   <<r<<ab;
      }
      return tmp.str();
}
string num2str(double r,int width) {
      ostringstream tmp;
      if (width <= 0) tmp<<r;
      else if (r == 0.0 || (abs(r) > 0.1 && abs(r)<pow(10.,max(width-3,1)))
	       || width <= 8) { 
	tmp<<fixed<<setw(max(width,5))<<setprecision(3)<<r; 
      } else { 
	tmp<<scientific<<setprecision(max(2,width-7))<<setw(max(9,width))<<r; 
      }
      return tmp.str();
}
string bool2str(bool b) {
  if (b) return "y";
  else return "n";
}
void printErr(string place, string message) {
  cout.setf(ios::internal);
  cerr<<" ("<<place<<":) "<<message<<"\n"; 
}
void printOut(string place, string message) {
  cout.setf(ios::internal);
  cout<<" ("<<setw(21)<<(place+":) ")<<message<<"\n"; 
}

//*********

// Natural cubic spline with vanishing 2nd order derivatives at boundaries.
// Given vectors x_i and y_i=f(x_i), the vector of 2nd derivs, y2 is returned.

void makeSpline(vector<double>& x, vector<double>& y, vector<double>& y2)
{
  int npoints(x.size());
  vector<double> u(npoints);
  y2[0]=u[0]=0.0;
  //Decomposition loop of tridiagonal alg. y2 and u used for tmp storage of 
  //decomposed factors.
  for (int i=1;i<npoints-1;i++ ) {
    double sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
    double p=sig*y2[i-1]+2.0;
    y2[i]=(sig-1.0)/p;
    u[i]=(y[i+1]-y[i])/(x[i+1]-x[i])-(y[i]-y[i-1])/(x[i]-x[i-1]);
    u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
  }
  y2[npoints-1]=0.0;
  for (int k=npoints-2;k>=0;k--) y2[k]=y2[k]*y2[k+1]+u[k];  
  return;
}

//Evaulation of spline, with values x and y, and splined 2nd der y2
double getSpline(const double x, vector<double>& xa, 
			      vector<double>& ya, vector<double>& y2a) {
  int npoints(xa.size());
  int k;
  double h,b,a;
  int klo=0;
  int khi=npoints-1;
  while (khi-klo > 1) {
    k=(khi+klo)>>1;
    if ( xa[k] > x ) khi=k;
    else klo=k;
  }
  h = xa[khi] - xa[klo];
  if ( h==0 ) cerr << " (getSpline:) Error: divide by zero\n" 
		 <<" khi ="<<khi<<" klo ="<<klo<<" : "
		 <<xa[khi]<<" - "<<xa[klo]<<endl;
  a = ( xa[khi] - x ) / h;
  b = ( x - xa[klo] ) / h;
  double y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
  return y;
}

//**************************************************************************

// External auxiliaries: extra four-products
  double m(const Vec4& v) {double s=m2(v); return (s >= 0)? sqrt(s):-sqrt(-s);}

double m2(const Vec4& v) {
  return pow2(v.e())-pow2(v.px())-pow2(v.py())-pow2(v.pz());}

double m2(const Vec4& v1, const Vec4& v2, const Vec4& v3) {
  double m2 = pow2(v1.e() + v2.e() + v3.e()) 
    - pow2(v1.px() + v2.px() + v3.px())
    - pow2(v1.py() + v2.py() + v3.py()) 
    - pow2(v1.pz() + v2.pz() + v3.pz());
  return m2; }  

double m2(const Vec4& v1, const Vec4& v2, 
		  const Vec4& v3, const Vec4& v4) {
  double m2 = pow2(v1.e() + v2.e() + v3.e() + v4.e()) 
    - pow2(v1.px() + v2.px() + v3.px() + v4.px())
    - pow2(v1.py() + v2.py() + v3.py() + v4.py()) 
    - pow2(v1.pz() + v2.pz() + v3.pz() + v4.pz());
  return m2; }  

double m2(const Particle& p1, const Particle& p2, const Particle& p3) {
  return m2(p1.p(),p2.p(),p3.p());}

// ******

// Convert Vec4 to primitive double array, for passing to Fortran.

void Vec4toD4(const Vec4& v, double* p) {
  p[0]=v.px();
  p[1]=v.py();
  p[2]=v.pz();
  p[3]=v.e();
}

  // MadGraph format, with energy first

  void Vec4toMG(const Vec4& v, double* p) {
    p[0]=v.e();
    p[1]=v.px();
    p[2]=v.py();
    p[3]=v.pz();
  }


//***********************************************************************

//*********

// Find all daughters of iAnc in event. Return a list with first color 
// daughter first, remaining partons unordered.

int findDaughters(Event& event,int iAnc,vector<int>& iDau) {

  // Sanity check
  if (iAnc <= 0 || iAnc >= event.size() ) return 0;
  iDau.resize(0);

  // Trace first daughter to the end.
  int iDa1=event[iAnc].daughter1();
  while (event[iDa1].status() < 0 && iDa1 > 0) iDa1=event[iDa1].daughter1();

  // Sanity check
  if (iDa1 == 0) return 0;

  // Store
  iDau.push_back(iDa1);

  // Add remaining partons in no particular order first
  for (int i=1;i<=event.size();i++) {
    if (i != iDa1 && event[i].isFinal() && isAncestor(event, i,iAnc)) {      
      iDau.push_back(i);
    }
  }

  // Return number of final-state daughters
  return iDau.size();

}

//*********

// Order a list of partons in color / branching sequence
// following the first parton in the list

int colorSort(Event& event,vector<int>& iList) {

  // Shorthand for final-state parton multiplicities
  int nPartons=iList.size();
  if (nPartons == 0) return 1;

  // Line up partons to follow first parton in color/branching order
  int iA=0;
  
  while (iA < nPartons-1) {

    int cType = event[iList[iA]].colType();

    // if iA is quark or gluon, find anti-color partner
    if (cType == 1 || cType == 2) {
      for (int iB=iA+1;iB<nPartons;iB++) {
	if (event[iList[iB]].acol() == event[iList[iA]].col()) {
	  int iTmp(iList[iA+1]);
	  iList[iA+1]=iList[iB];
	  iList[iB]=iTmp;
	  iA++;
	}
      }
    } 

    // if iA is antiquark, next is any quark of same flavor 
    else {
      int idA = event[iList[iA]].id();
      for (int iB=iA+1;iB<nPartons;iB++) {
	if (event[iList[iB]].id() == - idA) {
	  int iTmp(iList[iA+1]);
	  iList[iA+1]=iList[iB];
	  iList[iB]=iTmp;
	  iA++;
	}
      }
      
    }

  }

  // Normal return
  return 0;

}

//*********

// Check whether a given particle is an arbitrarily-steps-removed
// mother to another.
// (adapted for antenna showers, from Event::isAncestor from Pythia 8)
// (should probably add loop that stores additional mothers, so that we can 
// check whether all mothers are the same?)

bool isAncestor(Event& event, int i, int iAncestor) {

  // Begin loop to trace upwards from the daughter.
  int iUp = i;
  for ( ; ; ) {

    // If positive match then done.
    if (iUp == iAncestor) return true;

    // If out of range then failed to find match.
    if (iUp <= 0 || iUp > event.size()) return false;

    // Else go to mother1
    iUp = event[iUp].mother1(); 
    continue;

  }

  // End of loop. Should never reach beyond here.
  return false;

}

//**************************************************************************

// Math support auxiliaries

double Li2(const double x,const double kmax,const double xerr)
{
  if (x<0.0) return 0.5*Li2(x*x)-Li2(-x);
  if (x<=0.5)
    {
      double sum  = x;
      double term = x;
      for(int k=2; k<kmax; k++) 
	{
	  double rk = (k-1.0)/k;
	  term *= x;
	  term *= rk*rk;
	  sum += term;
	  if (fabs(term/sum) < xerr) return sum;
	}
      cout<<"Maximum number of iterations exceeded in Li2"<<endl;
      return sum;
    }
  if (x<1.0) 
    return M_PI*M_PI/6.0-Li2(1.0-x)-log(x)*log(1.0-x);
  if (x==1.0) 
    return M_PI*M_PI/6.0;
  if (x<=1.01)
    {
      const double eps = x - 1.0;
      const double lne = log(eps);
      const double c0 = M_PI*M_PI/6.0;
      const double c1 =   1.0 - lne;
      const double c2 = -(1.0 - 2.0*lne)/4.0;
      const double c3 =  (1.0 - 3.0*lne)/9.0;
      const double c4 = -(1.0 - 4.0*lne)/16.0;
      const double c5 =  (1.0 - 5.0*lne)/25.0;
      const double c6 = -(1.0 - 6.0*lne)/36.0;
      const double c7 =  (1.0 - 7.0*lne)/49.0;
      const double c8 = -(1.0 - 8.0*lne)/64.0;
      return c0+eps*(c1+eps*(c2+eps*(c3+eps*(c4+eps*(c5+eps*(c6+eps*(c7+eps*c8)))))));
    }
  double log_x=log(x);
  if (x<=2.0) 
    return M_PI*M_PI/6.0+Li2(1.0 - 1.0/x)-log_x*(log(1.0-1.0/x)+0.5*log_x);
  return M_PI*M_PI/3.0-Li2(1.0/x)-0.5*log_x*log_x;
}

int factorial(const int n) {
  int fac=1;
  for (int i=2;i<=n;i++) fac *= i;
  return fac;
}

int binomial(const int n, const int m) {
  if (m < 0 || m > n) return 0;
  else if (m == n || m == 0) return 1;
  else if (m == 1 || m == n-1) return n;
  else return factorial(n)/factorial(m)/factorial(n-m);
}

//*********

// The log of the Gamma function for positive x
// (Adapted from Numerical Recipes in Fortran, 1994)

double lnGamma(const double xx) 
{
  static double cof[6]={76.18009172947146,
			-86.50532032941677,
			24.01409824083091,
			-1.231739572450155,
			0.1208650973866179e-2,
			-0.5395239384953e-5};
  static double stp = 2.5066282746310005;
  if (xx <= 0) {
    printErr("lnGamma","cannot be called with x <= 0");
    return 0.0;
  }
  double y=xx;
  double tmp=xx+5.5;
  tmp = (xx+0.5)*log(tmp)-tmp;
  double ser=1.000000000190015;
  for (int j=0; j<6; j++) {
    y   += 1.0;
    ser += cof[j]/y;
  }
  return tmp + log(stp*ser/xx);
}

double Gamma(const double xx) 
{ 
  if (xx >= 0) {
    return exp(lnGamma(xx)); 
  } else {
    printErr("Gamma","Gamma(x<0) not implemented");
    return 0.0;
  }
}
  
//*********

// Solve the integral of fun1 over the interval [xlo,xhi]
// by adaptive gaussian quadrature (flat preweighting).
// Adapted from the CERNLIB DGAUSS routine by K.S. Kolbig. Error
// estimated by comparing 8-point and 16-point results.

double integrateGauss(double (*fun)(double), double xlo, double xhi, double tol)  
{
  //8-point unweighted
  static double x8[4]={0.96028985649753623,
		       0.79666647741362674,
		       0.52553240991632899, 
		       0.18343464249564980};
  static double w8[4]={0.10122853629037626,
		       0.22238103445337447,
		       0.31370664587788729,
		       0.36268378337836198};
  //16-point unweighted
  static double x16[8]={0.98940093499164993, 
		       0.94457502307323258, 
		       0.86563120238783174, 
		       0.75540440835500303, 
		       0.61787624440264375, 
		       0.45801677765722739,
		       0.28160355077925891, 
		       0.09501250983763744};
  static double w16[8]={0.027152459411754095,
		       0.062253523938647893,
		       0.095158511682492785,
		       0.12462897125553387,
		       0.14959598881657673,
		       0.16915651939500254,
		       0.18260341504492359,
		       0.18945061045506850};
  //boundary checks
  if (xlo == xhi) return 0.0;
  if ( xlo > xhi ) {
    cerr<<" (integrateGauss:) -> xhi < xlo"<<endl;
    return 0.0;
  }
  //initialize
  double sum = 0.0;
  double c = 0.001/abs(xhi-xlo);
  double zlo = xlo;
  double zhi = xhi;
    
  bool nextbin = true;
  
  while ( nextbin ) {
    
    double zmi = 0.5*(zhi+zlo); // midpoint
    double zmr = 0.5*(zhi-zlo); // midpoint, relative to zlo
    
    //calculate 8-point and 16-point quadratures
    double s8=0.0;
    for (int i=0;i<4;i++) {
      double dz = zmr * x8[i];
      s8 += w8[i]*(fun(zmi+dz) + fun(zmi-dz));
    }
    s8 *= zmr;
    double s16=0.0;
    for (int i=0;i<8;i++) {
      double dz = zmr * x16[i];
      s16 += w16[i]*(fun(zmi+dz) + fun(zmi-dz)); 
    }
    s16 *= zmr;
    if (abs(s16-s8) < tol*(1+abs(s16))) { 
      //precision in this bin OK, add to cumulative and go to next
      nextbin=true;
      sum += s16;
      //next bin: LO = end of current, HI = end of integration region.
      zlo=zhi;
      zhi=xhi;
      if ( zlo == zhi ) nextbin = false;
    } else {
      //precision in this bin not OK, subdivide.
      if (1.0 + c*abs(zmr) == 1.0) {
	cerr << " (integrateGauss:) too high accuracy required"<<endl;
	sum = 0.0 ;
	break;
      }
      zhi=zmi;
      nextbin=true;
    }
  }
  return sum;
}

//*********

// zbrent 
// solve fun(x)-r = 0 for x 
// tol is (probably?) an absolute tolerance parameter ...

double zbrent(double (*fun)(double), double r, double x1, double x2, double tol)
{
  int iter;
  double a=x1,b=x2,c=x2,d=x2-x1,e=x2-x1,min1,min2;
  double fa=fun(a)-r,fb=fun(b)-r,fc,p,q,r1,s,tol1,xm;
  double REALTINY=min(TINY,1e-12);
  tol = max(tol,REALTINY);

  // Check if there is a single zero in range
  if ( fa*fb > 0 ) { return 0.0; }

  // Start search
  fc=fb;
  for (iter=1;iter<max(1000,int(1.0/sqrt(tol)));iter++) {
    if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)) {
      c=a; 
      fc=fa;
      e=d=b-a;
    }
    if (abs(fc) < abs(fb)) {
      a=b;
      b=c;
      c=a;
      fa=fb;
      fb=fc;
      fc=fa;
    }
    tol1=2.0*REALTINY*fabs(b)+0.5*tol; 
    xm=0.5*(c-b);
    if (fabs(xm) <= tol1 || fb == 0.0) return b;
    if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) {
      s=fb/fa;
      if (a == c) {
	p=2.0*xm*s;
	q=1.0-s;
      } else {
	q=fa/fc;
	r1=fb/fc;
	p=s*(2.0*xm*q*(q-r1)-(b-a)*(r1-1.0));
	q=(q-1.0)*(r1-1.0)*(s-1.0);
      }
      if (p > 0.0) q = -q;
      p=fabs(p);
      min1=3.0*xm*q-fabs(tol1*q);
      min2=fabs(e*q);
      if (2.0*p < (min1 < min2 ? min1 : min2)) {
	e=d;
	d=p/q;
      } else {
	d=xm;
	e=d;
      }
    } else {
      d=xm;
      e=d;
    }
    a=b; 
    fa=fb;
    if (fabs(d) > tol1)
      b += d;
    else
      b += (xm > 1) ? tol1 : - tol1;
    fb=fun(b)-r;
  }
  cerr<<"(brent:) -> Maximum number of iterations exceeded"<<endl;
  return 0.0;
}

}
